import json, logging, os, shutil
from typing import Dict, List, Tuple
import tqdm
import torch
from torch.utils.data import DataLoader
from ednaml.config.EdnaMLConfig import EdnaMLConfig
from ednaml.crawlers import Crawler
from ednaml.models.ModelAbstract import ModelAbstract
from ednaml.utils.LabelMetadata import LabelMetadata


class BaseDeploy:
    """Base deployment class
    """
    model: ModelAbstract
    data_loader: DataLoader

    global_batch: int
    metadata: Dict[str, str]
    labelMetadata: LabelMetadata
    logger: logging.Logger

    def __init__(
        self,
        model: ModelAbstract,
        data_loader: DataLoader,
        epochs: int,
        logger: logging.Logger,
        crawler: Crawler,
        config: EdnaMLConfig,
        labels: LabelMetadata,
        **kwargs
    ):

        self.model = model
        self.parameter_groups = list(self.model.parameter_groups.keys())
       
        self.data_loader = data_loader
        self.logger = logger
        self.epochs = epochs
        self.global_epoch = 0
        self.global_batch = 0  # Current batch number in the epoch

        self.metadata = {}
        self.labelMetadata = labels
        self.crawler = crawler
        self.config = config

        self.buildMetadata(
            # TODO This is not gonna work with the torchvision wrapper -- ned to fix that; because crawler is not set up for that pattern...?
            crawler=crawler.classes,
            config=json.loads(config.export("json")),
        )

        self.model_is_built = False

    def buildMetadata(self, **kwargs):
        for keys in kwargs:
            self.metadata[keys] = kwargs.get(keys)

    def apply(
        self,
        step_verbose: int = 5,
        save_directory: str = "./checkpoint/",
        save_backup: bool = False,
        save_frequency: int = 1,
        backup_directory: str = None,
        gpus: int = 1,
        fp16: bool = False,
        model_save_name: str = None,
        logger_file: str = None,
    ):
        self.step_verbose = step_verbose
        self.save_directory = save_directory
        self.save_frequency = save_frequency
        self.backup_directory = None
        self.model_save_name = model_save_name
        self.logger_file = logger_file
        self.save_backup = save_backup
        if self.save_backup or self.config.SAVE.LOG_BACKUP:
            self.backup_directory = backup_directory
            os.makedirs(self.backup_directory, exist_ok=True)
        os.makedirs(self.save_directory, exist_ok=True)
        self.saveMetadata()

        self.gpus = gpus

        if self.gpus != 1:
            self.logger.warning("Multi-gpu or non-gpu not yet fully supported.")

        self.device = torch.device('cuda' if torch.cuda.is_available() else 'cpu')
        if self.gpus:
            #self.model.cuda() # moves the model into GPU
            self.logger.info("%i GPUs available"%self.gpus)
        self.model.to(self.device)

        self.fp16 = fp16
        # if self.fp16 and self.apex is not None:
        #    self.model, self.optimizer = self.apex.amp.initialize(self.model, self.optimizer, opt_level='O1')
        self.output_setup(**self.config.DEPLOYMENT.OUTPUT_ARGS)   # TODO 
    def saveMetadata(self):
        print("NOT saving metadata. saveMetadata() function not set up.")

    def deploy(self, continue_epoch=0, inference = False, ignore_plugins: List[str] = [], execute: bool = True, model_build: bool = None):
        if model_build or (model_build is None and not self.model_is_built):
            self.logger.info("Starting deployment")
            self.logger.info("Logging to:\t%s" % self.logger_file)
            if self.config.SAVE.LOG_BACKUP:
                self.logger.info(
                    "Logs will be backed up to drive directory:\t%s"
                    % self.backup_directory
                )
            
            #self.logger.info("Loading model from saved epoch %i" % (continue_epoch - 1))
            if continue_epoch > 0:
                load_epoch = continue_epoch - 1
                self.load(load_epoch, ignore_plugins=ignore_plugins)

            if inference:
                self.model.inference()
            else:
                self.model.eval()
            self.model_is_built = True
        else:
            if model_build is not None and not model_build:
                self.logger.info("Skipping model building due to `model_build=False`")
            elif self.model_is_built:
                self.logger.info("Skipping model building because model is already built. To force, set the `model_build` flag to True in `ed.deploy`")
            else:
                self.logger.info("Skipping model building")

        if execute:
            self.logger.info("Setting up plugin hooks. Plugins will fire during:  %s"%self.config.DEPLOYMENT.PLUGIN.HOOKS)
            self.model.set_plugin_hooks(self.config.DEPLOYMENT.PLUGIN.HOOKS)

            self.logger.info("Executing deployment for  %i epochs" % self.epochs)
            for epoch in range(self.epochs):
                self.logger.info("Starting epoch %i"%self.global_epoch)
                self.model.pre_epoch_hook(epoch=epoch)
                self.data_step()
                self.model.post_epoch_hook(epoch=epoch)
                self.global_epoch = epoch + 1
                if self.global_epoch % self.save_frequency == 0:
                    self.save()
                self.logger.info("Executing end of epoch steps")
                self.end_of_epoch(epoch=epoch)
            self.end_of_deployment()
            self.logger.info("Completed deployment task.")
        else:
            self.logger.info("Skipping execution of deployment task due to `execute = False`.")

    def data_step(self):   
        with torch.no_grad():
            for batch in tqdm.tqdm(
                self.data_loader, total=len(self.data_loader), leave=False
            ):    
                batch = self.move_to_device(batch)
                feature_logits, features, secondary_outputs = self.deploy_step(batch)

                self.output_step(feature_logits, features, secondary_outputs)
                # Log Metrics here and inside the model TODO
                self.global_batch += 1

    def move_to_device(self, batch) -> Tuple[torch.Tensor]:
        return (item.to(self.device) for item in batch)
        
    def deploy_step(self, batch):   # USER IMPLEMENTS
        batch = tuple(item.cuda() for item in batch)
        data, labels = batch    # TODO move plugins here to allow labels as well!!!!!!!!
        feature_logits, features, secondary_outputs = self.model(data)

        return feature_logits, features, secondary_outputs

    def end_of_epoch(self, epoch: int):
        pass
    
    def end_of_deployment(self):
        pass

    def output_setup(self, **kwargs): # USER IMPLEMENTS; kwargs from config.DEPLOYMENT.OUTPUT_ARGS
        self.logger.info("Warning: No output setup is performed")

    def output_step(self, logits, features, secondary): # USER IMPLEMENTS, ALSO, NEED SOME STEP LOGGING...????????
        if self.global_batch % self.config.LOGGING.STEP_VERBOSE == 0:
            self.logger.info("Warning: No output is generated at step %i"%self.global_batch)

    def load(self, load_epoch, ignore_plugins: List[str] = []):
        self.logger.info(
            "Loading a model from saved epoch %i"
            % (load_epoch)
        )
        model_load = self.model_save_name + "_epoch%i" % load_epoch + ".pth"

        if self.save_backup:
            self.logger.info(
                "Loading model from drive backup."
            )
            model_load_path = os.path.join(self.backup_directory, model_load)
        else:
            self.logger.info(
                "Loading model from local backup."
            )
            model_load_path = os.path.join(self.save_directory, model_load)

        self.model.load_state_dict(torch.load(model_load_path))
        self.logger.info(
            "Finished loading model state_dict from %s" % model_load_path
        )
        # Here, we will need to get a list of pickled or serialized plugin objects, then load them into a dictionary, then pass them into model
        # YES!!!!
        plugin_load = self.model_save_name + "_plugins.pth"

        if self.save_backup:
            self.logger.info(
                "Looking for model plugins from drive backup."
            )
            plugin_load_path = os.path.join(self.backup_directory, plugin_load)
        else:
            self.logger.info(
                "Looking for model plugins from local backup."
            )
            plugin_load_path = os.path.join(self.save_directory, plugin_load)

        if os.path.exists(plugin_load_path):
            
            self.model.loadPlugins(plugin_load_path, ignore_plugins=ignore_plugins)
            self.logger.info(
                "Loaded plugins from %s"%plugin_load_path
            )
        else:
            self.logger.info(
                "No plugins found at %s"%plugin_load_path
            )


    def save(self, save_epoch: int = None):
        if save_epoch is None:
            save_epoch = self.global_epoch
        self.logger.info("Performing save at epoch %i"%save_epoch)
        self.logger.info("Saving model plugins.")
        PLUGIN_SAVE_NAME = self.model_save_name + "_plugins.pth"
        # here, we save only the plugins...
        # Do not need model save, or training save, or log save...
        
        # A ModelAbstract can save its own plugins...?
        # No, we get back a plugin serialization object that we use to , well, save the plugin
        # Later, during plugin instantiation, we will need to write the code to load the plugin's save code...
        plugin_save = self.model.savePlugins()

        if len(plugin_save) > 0:
            torch.save(plugin_save, os.path.join(
                self.save_directory, PLUGIN_SAVE_NAME
            ))
            self.logger.info("Saved plugins: %s"%str(plugin_save.keys()))

            if self.save_backup:
                shutil.copy2(
                    os.path.join(self.save_directory, PLUGIN_SAVE_NAME),
                    self.backup_directory,
                )

                self.logger.info(
                    "Performing drive backup of model plugins"
                )
        else:
            self.logger.info("No plugins to save")

       
        


from ednaml.utils.blocks.InputAttention import InputAttention
from ednaml.utils.blocks.DenseAttention import DenseAttention
from ednaml.utils.blocks.ChannelAttention import ChannelAttention
from ednaml.utils.blocks.SpatialAttention import SpatialAttention
from ednaml.utils.blocks.ResnetBasicBlock import ResnetBasicBlock
from ednaml.utils.blocks.ResnetBottleneck import ResnetBottleneck
from ednaml.utils.blocks.ResnetInput import ResnetInput
from ednaml.utils.blocks.KMeansProxy import KMeansProxy



from typing import Any, List, Tuple, Union
import torchvision.transforms as T
from torch.utils.data import Dataset as TorchDataset
from ednaml.generators import Generator

import ednaml.utils
from ednaml.utils.LabelMetadata import LabelMetadata


class TextGenerator(Generator):
    """Base class for text generators"""

    def __init__(
        self, logger = None, gpus: int = 1, transforms={}, mode: str = "train", **kwargs
    ):
        """Initializes the Generator and builds the data transformer

        Args:
            gpus (_type_): _description_
            i_shape (_type_): _description_
            normalization_mean (_type_): _description_
            normalization_std (_type_): _description_
            normalization_scale (_type_): _description_
        """
        super().__init__(logger, gpus=gpus, transforms=transforms, mode=mode, **kwargs)

    def buildDataset(
        self, datacrawler, mode: str, transform: List[object], **kwargs
    ) -> TorchDataset:
        """Given the datacrawler with all the data, and the mode (could be
        any user-defined mode such as 'train', 'test', 'zsl', 'gzsl', etc), as
        as well the transform, return a TorchDataset

        Args:
            datacrawler (_type_): _description_
            mode (_type_): _description_
            transform (_type_): _description_

        Raises:
            NotImplementedError: _description_
        """
        raise NotImplementedError()

    def buildDataLoader(self, dataset, mode, batch_size, **kwargs):
        """Given a Torch Dataset, build a TorchDataloader to return a tensor, possibly with a collate function

        Args:
            dataset (_type_): _description_
            batch_size (_type_): _description_

        Raises:
            NotImplementedError: _description_
        """

        raise NotImplementedError()



from ednaml.models.ModelAbstract import ModelAbstract
from ednaml.plugins import ModelPlugin
import torch
import torch.nn.functional
from sklearn.neighbors import KDTree

class KMeansProxy(ModelPlugin):
    name = "KMeansProxy"
    def __init__(self, num_clusters=10, dimensions=768, dist="euclidean", rand_seed=145845, epochs=3, **kwargs):
        super().__init__(num_clusters = num_clusters, dimensions=dimensions, dist=dist, rand_seed = rand_seed, epochs=epochs)


    def build_plugin(self, **kwargs):
        self.num_clusters = kwargs.get("num_clusters")
        self.rand_seed = kwargs.get("rand_seed", 145845)
        self.dimensions = kwargs.get("dimensions")
        self.dist = kwargs.get("dist", "euclidean")    # TODO
        self.epochs = kwargs.get("epochs", 3)
        self.epoch_count = 0
        self.pre_epoch_flag = False
        self.pre_epoch_num = -1
        self.post_epoch_num = -1
        self.activated = False
        self.kdcluster = None
        self._dist_func = self.build_dist(self.dist)
        if self.dist == "euclidean":
            self._preprocess = lambda x:x
        elif self.dist == "cosine":
            self._preprocess = torch.nn.functional.normalize

    def build_dist(self, dist="euclidean"):
        if dist == "euclidean":
            return self.l2dist
        elif dist == "cosine":
            return self.cosdist
        else:
            raise ValueError("Invalid value for dist: %s"%dist)

    def l2dist(self, x):
        return torch.argmin(torch.sqrt(((self.cluster_means - x)**2).sum(1)))
    def cosdist(self, x):   #https://stackoverflow.com/questions/46409846/using-k-means-with-cosine-similarity-python
        return self.l2dist(torch.nn.functional.normalize(x, dim=0))
        
    
    def build_params(self, **kwargs):
        self.cluster_means = torch.rand(self.num_clusters, self.dimensions)
        self.cluster_counts = torch.zeros(self.num_clusters)


    def post_forward(self, x, feature_logits, features, secondary_outputs, model, **kwargs):
        if not self.activated:
            # perform the training here
            self.update_minibatch_kmeans_clusters(features)
            
            return feature_logits, features, secondary_outputs, kwargs, {}
        else:
            dist, labels = self.compute_labels(features)
            return feature_logits, features, secondary_outputs, kwargs, {"cluster_mean": labels, "distance": dist}

    def post_epoch(self, model: ModelAbstract, epoch: int = 0, **kwargs):
        if not self.activated:  # If already activated, we do not need to change anything
            self.post_epoch_num = epoch    # Maybe have a better check in case something drops or epochs get skipped...
            if self.pre_epoch_flag: # Means we had a pre_epoch flag so an epoch was completed.
                self.pre_epoch_flag = False # reset the flag
                self.epoch_count+=1
                if self.post_epoch_num != self.pre_epoch_num:
                    raise ValueError("Epoch may have been skipped. Before: %i\tAfter: %i"%(self.pre_epoch_num, self.post_epoch_num))
            self.activated = self.epoch_count > self.epochs # better way to deal with this whole situation, i.e. once activated, never changes...!
        if self.activated and self.kdcluster is None:
            self._logger.info("KMP is active. Generating kdcluster")
            if self.dist == "euclidean":
                self.kdcluster = KDTree(self.cluster_means)
            elif self.dist == "cosine":
                self.kdcluster = KDTree(torch.nn.functional.normalize(self.cluster_means))
            else:
                raise ValueError("Invalid value for dist: %s"%self.dist)

            #model.classifier(self.cluster_means.cuda())
            with torch.no_grad():
                self.labels = torch.argmax(model.classifier(self.cluster_means.cuda()), dim=1).cpu()


    def pre_epoch(self, model: ModelAbstract, epoch: int = 0, **kwargs):
        self.pre_epoch_flag = True
        self.pre_epoch_num = epoch

    def update_minibatch_kmeans_clusters(self, batch: torch.Tensor, ):
        #batch_device = batch.get_device()   # TODO fix this hacky with the hack from https://stackoverflow.com/questions/58926054/how-to-get-the-device-type-of-a-pytorch-module-conveniently 2nd answer
        local_batch = batch.cpu()
        V = torch.zeros(self.num_clusters)
        idxs = torch.empty(batch.shape[0], dtype=torch.int)
        for j, x in enumerate(local_batch):
            idxs[j] = self._dist_func(x)

        # Update centers:
        for j, x in enumerate(local_batch):
            V[idxs[j]] += 1
            eta = 1.0 / V[idxs[j]]
            self.cluster_means[idxs[j]] = (1.0 - eta) * self.cluster_means[idxs[j]] + eta * x

    def compute_labels(self, batch):
        """Compute the cluster labels for dataset X given centers C.
        """
        # labels = np.argmin(pairwise_distances(C, X), axis=0) # THIS REQUIRES TOO MUCH MEMORY FOR LARGE X
        q_batch = batch.cpu()
        dist, idx = self.kdcluster.query(self._preprocess(q_batch), k=1, return_distance=True)   #.squeeze()
        # TODO convert idx to the actual cluster means to the actual cluster labels...
        return torch.from_numpy(dist).squeeze(1), torch.stack([self.labels[item[0]] for item in idx])



"""cluster_distances = torch.zeros(self.num_clusters)
    for cluster in range(self.num_clusters):
        cluster_distances[cluster] = sum(torch.sqrt((features - self.cluster_means[cluster])**2))
    c = torch.argmin(cluster_distances)
    self.cluster_counts[c] += 1
    self.cluster_means[c] += 1.0/self.cluster_counts[c]*(features - self.cluster_means[c])

"""


import mmap
from types import FunctionType



class IterableFile:

    def __init__(self, file_path, chunk_size = 50000, line_callback: FunctionType = None):
        # TODO use chunk-size to preload some chunk into memory for ammortized overhead?
        self.row_count = 0
        for _ in (row for row in open(file_path, "r")):
            self.row_count+=1
        self.file_path = file_path
        self.line_callback = line_callback

    def __len__(self):
        return self.row_count

    def __iter__(self):
        return (self.line_callback(row) for row in open(self.file_path, "r"))


"""
self.file_obj = open(file_path, "r+b")

        try:    # UNIX vs Windows
            s = mmap.PROT_READ
        except AttributeError:
            s = mmap.ACCESS_READ

        self.mmap = mmap.mmap(self.file_obj.fileno(), 0, prot=s)
        self.mmap_iter = iter(self.mmap.readline, b"")
        
"""


from typing import Dict

from ednaml.config import BaseConfig


class ExecutionDatareaderConfig(BaseConfig):
    DATAREADER: str
    CRAWLER_ARGS: Dict[str, str]
    DATASET_ARGS: Dict[str, str]
    GENERATOR: str
    GENERATOR_ARGS: Dict[str, str]
    

    def __init__(self, datareader_dict):
        self.DATAREADER = datareader_dict.get("DATAREADER", "DataReader")
        self.CRAWLER_ARGS = datareader_dict.get("CRAWLER_ARGS", {})
        self.DATASET_ARGS = datareader_dict.get("DATASET_ARGS", {})
        self.GENERATOR = datareader_dict.get(
            "GENERATOR", None
        )
        self.GENERATOR_ARGS = datareader_dict.get("GENERATOR_ARGS", {})



from ednaml.generators import Generator
from ednaml.utils.LabelMetadata import LabelMetadata
from typing import List, Tuple, Union
import torchvision.transforms as T
from torch.utils.data import Dataset as TorchDataset
import ednaml.utils


class ImageGenerator(Generator):
    """Base class for image dataset generators"""

    def __init__(
        self, logger=None, gpus: int = 1, transforms={}, mode: str = "train", **kwargs
    ):
        """Initializes the Generator and builds the data transformer

        Args:
            gpus (_type_): _description_
            i_shape (_type_): _description_
            normalization_mean (_type_): _description_
            normalization_std (_type_): _description_
            normalization_scale (_type_): _description_
        """
        super().__init__(logger, gpus=gpus, transforms=transforms, mode=mode, **kwargs)

    def build_transforms(self, transforms, mode, **kwargs):
        return T.Compose(self._build_transforms(**transforms.ARGS))
        

    def _build_transforms(
        self,
        i_shape: Union[List[int], Tuple[int, int]],
        channels: int,
        normalization_mean: float,
        normalization_std: float,
        normalization_scale: float,
        **kwargs
    ) -> List[object]:
        """Builds the transforms for the images in dataset. This can be replaced for custom set of transforms

        Args:
            i_shape (Union[List[int,int],Tuple[int,int]]): _description_
            normalization_mean (float): _description_
            normalization_std (float): _description_
            normalization_scale (float): _description_

        Returns:
            _type_: _description_
        """
        if(i_shape == [] and channels == 0 and normalization_mean == 0 and normalization_std == 0 and normalization_scale == 0):
            return []
        
        (
            normalization_mean,
            normalization_std,
        ) = ednaml.utils.extend_mean_arguments(
            [normalization_mean, normalization_std], channels
        )
        transformer_primitive = []
        transformer_primitive.append(T.Resize(size=i_shape))
        if kwargs.get("h_flip", 0) > 0:
            transformer_primitive.append(
                T.RandomHorizontalFlip(p=kwargs.get("h_flip"))
            )
        if kwargs.get("t_crop", False):
            transformer_primitive.append(T.RandomCrop(size=i_shape))
        transformer_primitive.append(T.ToTensor())
        transformer_primitive.append(
            T.Normalize(mean=normalization_mean, std=normalization_std)
        )
        if kwargs.get("rea", False):
            transformer_primitive.append(
                T.RandomErasing(
                    p=0.5, scale=(0.02, 0.4), value=kwargs.get("rea_value", 0)
                )
            )
        return transformer_primitive

    def buildDataset(
        self, datacrawler, mode: str, transform: List[object], **kwargs
    ) -> TorchDataset:
        """Given the datacrawler with all the data, and the mode (could be
        any user-defined mode such as 'train', 'test', 'zsl', 'gzsl', etc), as
        as well the transform, return a TorchDataset

        Args:
            datacrawler (_type_): _description_
            mode (_type_): _description_
            transform (_type_): _description_

        Raises:
            NotImplementedError: _description_
        """
        raise NotImplementedError()

    def buildDataLoader(self, dataset, mode, batch_size, **kwargs):
        """Given a Torch Dataset, build a TorchDataloader to return a tensor, possibly with a collate function

        Args:
            dataset (_type_): _description_
            batch_size (_type_): _description_

        Raises:
            NotImplementedError: _description_
        """

        raise NotImplementedError()

    def getNumEntities(self, datacrawler, mode, **kwargs) -> LabelMetadata:
        """Return the number of classes in the overall label scheme
        Raises:
            NotImplementedError: _description_
        """

        raise NotImplementedError()



import json, logging, os, shutil
from logging import Logger
from typing import Dict, List, Tuple
from ednaml.config.StorageConfig import StorageConfig

import torch
from torch.utils.data import DataLoader
import ednaml.loss.builders
from ednaml.config.EdnaMLConfig import EdnaMLConfig
from ednaml.crawlers import Crawler
from ednaml.models.ModelAbstract import ModelAbstract
from ednaml.utils.LabelMetadata import LabelMetadata
from ednaml.storage.BaseStorage import BaseStorage

class BaseTrainer:
    model: ModelAbstract
    loss_fn: Dict[
        str, ednaml.loss.builders.LossBuilder
    ]  # output-name: lossBuilder
    optimizer: Dict[str, torch.optim.Optimizer]
    loss_optimizer = Dict[str, List[torch.optim.Optimizer]]
    scheduler: Dict[str, torch.optim.lr_scheduler._LRScheduler]
    loss_scheduler: Dict[str, List[torch.optim.lr_scheduler._LRScheduler]]

    skipeval: bool
    train_loader: DataLoader
    test_loader: DataLoader

    epochs: int
    logger: Logger
    global_batch: int
    global_epoch: int
    num_losses: int
    losses: Dict[str, List[int]]
    metadata: Dict[str, str]
    labelMetadata: LabelMetadata
    logger: logging.Logger
    storage: BaseStorage

    def __init__(
        self,
        model: ModelAbstract,
        loss_fn: List[ednaml.loss.builders.LossBuilder],
        optimizer: List[torch.optim.Optimizer],
        loss_optimizer: List[torch.optim.Optimizer],
        scheduler: List[torch.optim.lr_scheduler._LRScheduler],
        loss_scheduler: List[torch.optim.lr_scheduler._LRScheduler],
        train_loader: DataLoader,
        test_loader: DataLoader,
        epochs: int,
        skipeval: bool,
        logger: Logger,
        crawler: Crawler,
        config: EdnaMLConfig,
        labels: LabelMetadata,
        storage: BaseStorage,
        **kwargs
    ):
        self.model = model
        self.parameter_groups = list(self.model.parameter_groups.keys())
        self.loss_fn_order = {
            idx: lossbuilder.loss_labelname
            for idx, lossbuilder in enumerate(loss_fn)
        }
        self.loss_fn = {
            lossbuilder.loss_labelname: lossbuilder for lossbuilder in loss_fn
        }
        self.num_losses = len(self.loss_fn)
        self.losses = {lossname: [] for lossname in self.loss_fn}
        self.loss_optimizer = {
            self.loss_fn_order[idx]: loss_optimizer_content
            for idx, loss_optimizer_content in enumerate(loss_optimizer)
        }
        if type(loss_scheduler) is list:
            self.loss_scheduler = {
                self.loss_fn_order[idx]: loss_scheduler_content
                for idx, loss_scheduler_content in enumerate(loss_scheduler)
            }
        else:
            self.loss_scheduler = {
                self.loss_fn_order[idx]: loss_scheduler
                for idx in range(self.num_losses)
            }

        self.optimizer = {
            self.parameter_groups[idx]: optimizer_item
            for idx, optimizer_item in enumerate(optimizer)
        }
        self.scheduler = {
            self.parameter_groups[idx]: scheduler_item
            for idx, scheduler_item in enumerate(scheduler)
        }
        self.skipeval = skipeval
        self.train_loader = train_loader
        self.test_loader = test_loader

        self.epochs = epochs
        self.logger = logger

        self.global_batch = 0  # Current batch number in the epoch
        self.global_epoch = 0

        self.metadata = {}
        self.labelMetadata = labels
        self.crawler = crawler
        self.config = config
        self.storage = storage
        # Add later -- download data from Azure/other file instance
        #self.downloadData()

        self.buildMetadata(
            # TODO This is not gonna work with the torchvision wrapper -- ned to fix that; because crawler is not set up for that pattern...?
            crawler=crawler.classes,
            config=json.loads(config.export("json")),
        )

        self.accumulation_steps = kwargs.get("accumulation_steps")
        self.accumulation_count = 0
        self.evaluateFlag = False
        self.saveFlag = False
        self.init_setup(**kwargs)

    def init_setup(self, **kwargs):
        pass

    def downloadData(self): #TODO
        #storageinstance = AzureStorage(self.cfg.STORAGE)
        self.storage.read()


    def buildMetadata(self, **kwargs):
        for keys in kwargs:
            self.metadata[keys] = kwargs.get(keys)

    def apply(
        self,
        step_verbose: int = 5,
        save_frequency: int = 5,
        test_frequency: int = 5,
        save_directory: str = "./checkpoint/",
        save_backup: bool = False,
        backup_directory: str = None,
        gpus: int = 1,
        fp16: bool = False,
        model_save_name: str = None,
        logger_file: str = None,
    ):
        self.step_verbose = step_verbose
        self.save_frequency = save_frequency
        self.test_frequency = test_frequency
        self.save_directory = save_directory
        self.backup_directory = None
        self.model_save_name = model_save_name
        self.logger_file = logger_file
        self.save_backup = save_backup
        if self.save_backup or self.config.SAVE.LOG_BACKUP:
            self.backup_directory = backup_directory
            os.makedirs(self.backup_directory, exist_ok=True)
        os.makedirs(self.save_directory, exist_ok=True)
        self.saveMetadata()

        self.gpus = gpus

        if self.gpus != 1:
            self.logger.warning("Multi-gpu or non-gpu not yet fully supported.")


        self.device = torch.device('cuda' if torch.cuda.is_available() else 'cpu')
        if self.gpus:
            #self.model.cuda() # moves the model into GPU
            self.logger.info("%i GPUs available"%self.gpus)
        self.model.to(self.device)

        self.fp16 = fp16
        # if self.fp16 and self.apex is not None:
        #    self.model, self.optimizer = self.apex.amp.initialize(self.model, self.optimizer, opt_level='O1')

    def saveMetadata(self):
        print("NOT saving metadata. saveMetadata() function not set up.")

    def save(self, save_epoch: int = None):
        if save_epoch is None:
            save_epoch = self.global_epoch
        self.logger.info("Saving model, optimizer, and scheduler.")
        MODEL_SAVE = self.model_save_name + "_epoch%i" % save_epoch + ".pth"
        TRAINING_SAVE = (
            self.model_save_name + "_epoch%i" % save_epoch + "_training.pth"
        )

        save_dict = {}
        save_dict["optimizer"] = {
            pgn: self.optimizer[pgn].state_dict()
            for pgn in self.parameter_groups
        }
        save_dict["scheduler"] = {
            pgn: self.scheduler[pgn].state_dict()
            for pgn in self.parameter_groups
        }
        save_dict["loss_fn"] = {
            lossname: self.loss_fn[lossname].state_dict()
            for lossname in self.loss_fn
        }
        save_dict["loss_optimizer"] = {
            lossname: (
                self.loss_optimizer[lossname].state_dict()
                if self.loss_optimizer[lossname] is not None
                else None
            )
            for lossname in self.loss_optimizer
        }
        save_dict["loss_scheduler"] = {
            lossname: (
                self.loss_scheduler[lossname].state_dict()
                if self.loss_scheduler[lossname] is not None
                else None
            )
            for lossname in self.loss_scheduler
        }
        # TODO need to check if ModelAbstract contains the plugins that came with it...
        # if not, we will need to save plugins in a plugins artifact.
        # And even if they did, we need to split up model and plugins, since they are not strictly part of the model.

        # save_dict["loss_optimizer"] = [self.loss_optimizer[idx].state_dict() if self.loss_optimizer[idx] is not None else None for idx in range(self.num_losses)]
        # save_dict["loss_scheduler"] = [self.loss_scheduler[idx].state_dict() if self.loss_scheduler[idx] is not None else None for idx in range(self.num_losses)]
        
        # TODO
        # Have a Save Class that stores methods, 
        # .save(object)

        torch.save(
            self.model.state_dict(),
            os.path.join(self.save_directory, MODEL_SAVE),
        )
        torch.save(save_dict, os.path.join(self.save_directory, TRAINING_SAVE))

        if self.save_backup:
            shutil.copy2(
                os.path.join(self.save_directory, MODEL_SAVE),
                self.backup_directory,
            )
            shutil.copy2(
                os.path.join(self.save_directory, TRAINING_SAVE),
                self.backup_directory,
            )

            self.logger.info(
                "Performing drive backup of model, optimizer, and scheduler."
            )
        
        if self.config.SAVE.LOG_BACKUP:
            LOGGER_SAVE = os.path.join(self.backup_directory, self.logger_file)
            #appending the logs to azure
            #self.storage.copy(os.path.join(self.save_directory, self.logger_file)) # TODO
            # This is the local path where we are copying data . This can be commented/removed later
            if os.path.exists(LOGGER_SAVE):
                os.remove(LOGGER_SAVE)
            shutil.copy2(
                os.path.join(self.save_directory, self.logger_file), LOGGER_SAVE
            )

                

    def load(self, load_epoch):
        self.logger.info(
            "Resuming training from epoch %i. Loading saved state from %i"
            % (load_epoch + 1, load_epoch)
        )
        model_load = self.model_save_name + "_epoch%i" % load_epoch + ".pth"
        training_load = (
            self.model_save_name + "_epoch%i" % load_epoch + "_training.pth"
        )

        if self.save_backup:
            self.logger.info(
                "Loading model, optimizer, and scheduler from drive backup."
            )
            model_load_path = os.path.join(self.backup_directory, model_load)
            training_load_path = os.path.join(
                self.backup_directory, training_load
            )

        else:
            self.logger.info(
                "Loading model, optimizer, and scheduler from local backup."
            )
            model_load_path = os.path.join(self.save_directory, model_load)
            training_load_path = os.path.join(
                self.save_directory, training_load
            )

        self.model.load_state_dict(torch.load(model_load_path))
        self.logger.info(
            "Finished loading model state_dict from %s" % model_load_path
        )

        checkpoint = torch.load(training_load_path)
        for pgn in self.parameter_groups:
            self.optimizer[pgn].load_state_dict(checkpoint["optimizer"][pgn])

            self.scheduler[pgn].load_state_dict(checkpoint["scheduler"][pgn])
        self.logger.info(
            "Finished loading optimizer state_dict from %s" % training_load_path
        )
        self.logger.info(
            "Finished loading scheduler state_dict from %s" % training_load_path
        )

        for lossname in self.loss_fn:
            self.loss_fn[lossname].load_state_dict(
                checkpoint["loss_fn"][lossname]
            )
            if self.loss_optimizer[lossname] is not None:
                self.loss_optimizer[lossname].load_state_dict(
                    checkpoint["loss_optimizer"][lossname]
                )
            if self.loss_scheduler[lossname] is not None:
                self.loss_scheduler[lossname].load_state_dict(
                    checkpoint["loss_scheduler"][lossname]
                )

            self.logger.info(
                "Finished loading loss state_dict from %s" % training_load_path
            )

        # for idx in range(self.num_losses):
        #    if self.loss_optimizer[idx] is not None:
        #        self.loss_optimizer[idx].load_state_dict(checkpoint["loss_optimizer"][idx])
        #    if self.loss_scheduler[idx] is not None:
        #        self.loss_scheduler[idx].load_state_dict(checkpoint["loss_scheduler"][idx])

    def train(self, continue_epoch=0):
        self.logger.info("Starting training")
        self.logger.info("Logging to:\t%s" % self.logger_file)
        self.logger.info(
            "Models will be saved to local directory:\t%s" % self.save_directory
        )
        if self.config.SAVE.LOG_BACKUP:
            self.logger.info(
                "Logs will be backed up to drive directory:\t%s"
                % self.backup_directory
            )
        if self.save_backup:
            self.logger.info(
                "Models will be backed up to drive directory:\t%s"
                % self.backup_directory
            )
        self.logger.info(
            "Models will be saved with base name:\t%s_epoch[].pth"
            % self.model_save_name
        )
        self.logger.info(
            "Optimizers will be saved with base name:\t%s_epoch[]_optimizer.pth"
            % self.model_save_name
        )
        self.logger.info(
            "Schedulers will be saved with base name:\t%s_epoch[]_scheduler.pth"
            % self.model_save_name
        )

        if continue_epoch > 0:
            load_epoch = continue_epoch - 1
            self.load(load_epoch)

        if not self.skipeval:
            self.logger.info("Performing initial evaluation...")
            self.initial_evaluate()
        else:
            self.logger.info("Skipping initial evaluation.")
        
        self.logger.info("Starting training from %i" % continue_epoch)
        self.zeroGrad()
        self.evaluateFlag = False
        self.saveFlag = False
        for epoch in range(self.epochs + 1):
            if epoch >= continue_epoch:
                # TODO pre epoch
                self.epoch_step(epoch)
                # TODO post epoch
            else:
                self.global_epoch = epoch + 1

        if self.evaluateFlag:
            self.logger.info("Final: Evaluating model at test-frequency")
            self.evaluate()
            self.evaluateFlag = False
        if self.saveFlag:
            self.logger.info("Final: Saving model at save-frequency")
            self.save(self.global_epoch - 1)
            self.saveFlag = False

    def initial_evaluate(self):
        """Evaluation of model before we start training"""
        self.evaluate()

    def epoch_step(self, epoch):
        """Trains model for an epoch."""
        for batch in self.train_loader:
            if self.global_batch == 0:
                self.printOptimizerLearningRates()

            self.model.train() #train == we are tracking all numbers and computation graph
            batch = self.move_to_device(batch)
            loss: torch.Tensor = self.step(batch) #perform function and returns loss
            loss = loss / self.accumulation_steps
            loss.backward()
            self.accumulation_count += 1

            if self.accumulation_count % self.accumulation_steps == 0:
                self.updateGradients()
                self.accumulation_count = 0
                if self.evaluateFlag:
                    self.logger.info("Evaluating model at test-frequency")
                    self.evaluate()
                    self.evaluateFlag = False
                if self.saveFlag:
                    self.logger.info("Saving model at save-frequency")
                    self.save(self.global_epoch - 1)
                    self.saveFlag = False

            self.global_batch += 1
            if (self.global_batch + 1) % self.step_verbose == 0:
                self.printStepInformation()

        self.global_batch = 0
        self.stepSchedulers()
        self.stepLossSchedulers()

        self.logger.info(
            "{0} Completed epoch {1} {2}".format(
                "*" * 10, self.global_epoch, "*" * 10
            )
        )

        if self.global_epoch % self.test_frequency == 0:
            if self.accumulation_steps > 0:
                self.logger.info(
                    "Model evaluation triggered, but gradients still need"
                    " accumulation. Will evaluate after accumulation."
                )
                self.evaluateFlag = True
            else:
                self.evaluate()

        if self.global_epoch % self.save_frequency == 0:
            if self.accumulation_steps > 0:
                self.logger.info(
                    "Model save triggered, but gradients still need"
                    " accumulation. Will save after accumulation."
                )
                self.saveFlag = True
            else:
                self.save()
        self.global_epoch += 1
    
    def move_to_device(self, batch) -> Tuple[torch.Tensor]:
        return (item.to(self.device) for item in batch)

    def step(self, batch) -> torch.Tensor:
        # compute the loss, and return it
        # print("!!!!!!!!!! batch !!!!!!!!!!!!!!!",batch)

        raise NotImplementedError()

    def updateGradients(self):
        self.stepOptimizers()
        self.stepLossOptimizers()
        self.zeroGrad()

    def zeroGrad(self):
        self.zeroGradOptimizers()
        self.zeroGradLossOptimizers()

    def printStepInformation(self):
        loss_avg = 0.0
        for lossname in self.losses:
            loss_avg += (
                sum(self.losses[lossname][-self.step_verbose :])
                / self.step_verbose
            )
        loss_avg /= self.num_losses
        if(len(self.softaccuracy[-100:]) > 0):
            soft_avg = sum(self.softaccuracy[-100:]) / float(
                len(self.softaccuracy[-100:]) 
            )
        else:
            soft_avg= 0
        self.logger.info(
            "Epoch{0}.{1}\tTotal Avg Loss: {2:.3f} Softmax: {3:.3f}".format(
                self.global_epoch, self.global_batch, loss_avg, soft_avg
            )
        )

    def evaluate(self):
        self.model.eval()
        logit_labels, true_labels, features = self.evaluate_impl()
        return logit_labels, true_labels, self.crawler.classes, features

    def evaluate_impl(self):
        raise NotImplementedError()

    def zeroGradOptimizers(self):
        for optim in self.optimizer:
            self.optimizer[optim].zero_grad()

    def zeroGradLossOptimizers(self):
        for lossname in self.loss_fn:
            if self.loss_optimizer[lossname] is not None:
                self.loss_optimizer[lossname].zero_grad()

    def stepOptimizers(self):
        for optim in self.optimizer:
            self.optimizer[optim].step()

    def stepLossOptimizers(self):
        for lossname in self.loss_fn:
            if (
                self.loss_optimizer[lossname] is not None
            ):  # In case loss object doesn;t have any parameters, this will be None. See optimizers.StandardLossOptimizer
                self.loss_optimizer[lossname].step()

    def stepSchedulers(self):
        for scheduler in self.scheduler:
            self.scheduler[scheduler].step()

    def stepLossSchedulers(self):
        for lossname in self.loss_fn:
            if self.loss_scheduler[lossname] is not None:
                self.loss_scheduler[lossname].step()

    def printOptimizerLearningRates(self):
        for param_group_name in self.optimizer:
            lrs = self.scheduler[param_group_name].get_last_lr()
            lrs = sum(lrs) / float(len(lrs))
            self.logger.info(
                "Parameter Group `{0}`: Starting epoch {1} with {2} steps and"
                " learning rate {3:2.5E}".format(
                    param_group_name,
                    self.global_epoch,
                    len(self.train_loader) - (len(self.train_loader) % 10),
                    lrs,
                )
            )



import inspect, os

import ednaml


REGISTERED_EDNA_COMPONENTS = {}


def register(func, functype):
    try:
        fname = inspect.getfile(func)
    except TypeError:
        fname = os.path.abspath(func.__module__)
    print("Registering a %s: %s, from file: %s"%(functype, str(func), fname))
    if fname not in REGISTERED_EDNA_COMPONENTS:
        REGISTERED_EDNA_COMPONENTS[fname] = {}
    # TODO also add in some type of lookup for the file basename...?
    if functype is "model_plugin":
        if "model_plugin" not in REGISTERED_EDNA_COMPONENTS[fname]:
            REGISTERED_EDNA_COMPONENTS[fname]["model_plugin"] = []
        REGISTERED_EDNA_COMPONENTS[fname]["model_plugin"].append(func)
    else:
        REGISTERED_EDNA_COMPONENTS[fname][functype] = func


def register_crawler(func):
    register(func, "crawler")
    return func

def register_model(func):
    register(func, "model")
    return func

def register_generator(func):
    register(func, "generator")
    return func

def register_storage(func):
    register(func, "storage")
    return func

def register_trainer(func):
    register(func, "trainer")
    return func

def register_deployment(func):
    register(func, "deployment")
    return func


def register_model_plugin(func):
    register(func, "model_plugin")
    return func

def register_model_pre_plugin(func):
    pass



from ednaml.config import BaseConfig
from ednaml.config.ConfigDefaults import ConfigDefaults
from ednaml.config.ExecutionDatareaderConfig import ExecutionDatareaderConfig
from ednaml.config.ExecutionPluginConfig import ExecutionPluginConfig
from typing import Dict

class DeploymentConfig(BaseConfig):
    OUTPUT_ARGS: Dict[str,str]
    DEPLOYMENT_ARGS: Dict[str,str]
    DEPLOY: str
    DATAREADER: ExecutionDatareaderConfig
    EPOCHS: int
    PLUGIN: ExecutionPluginConfig


    def __init__(self, deployment_dict, defaults: ConfigDefaults):
        self.OUTPUT_ARGS = deployment_dict.get(
            "OUTPUT_ARGS", defaults.OUTPUT_ARGS
        )
        self.DEPLOYMENT_ARGS = deployment_dict.get(
            "DEPLOYMENT_ARGS", defaults.DEPLOYMENT_ARGS
        )
        self.DEPLOY = deployment_dict.get(
            "DEPLOY", defaults.DEPLOY
        )
        self.DATAREADER = ExecutionDatareaderConfig(
            deployment_dict.get("DATAREADER", {})
        )
        self.EPOCHS = deployment_dict.get("EPOCHS", defaults.DEPLOYMENT_EPOCHS)
        self.PLUGIN = ExecutionPluginConfig(deployment_dict.get("PLUGIN", {}))


import unicodedata
import collections

SPIECE_UNDERLINE = "▁"


def preprocess_text(inputs, remove_space=True, do_lower_case=True):
    if remove_space:
        outputs = " ".join(inputs.strip().split())
    else:
        outputs = inputs
    outputs = outputs.replace("``", '"').replace("''", '"')
    outputs = unicodedata.normalize("NFKD", outputs)
    outputs = "".join([c for c in outputs if not unicodedata.combining(c)])
    if do_lower_case:
        outputs = outputs.lower()
    return outputs


def encode_pieces(sp_model, text, return_unicode=True, sample=False):
    """turn sentences into word pieces."""
    text = preprocess_text(
        text,
    )
    if not sample:
        pieces = sp_model.EncodeAsPieces(text)
    else:
        pieces = sp_model.SampleEncodeAsPieces(text, 64, 0.1)
    new_pieces = []
    for piece in pieces:
        if len(piece) > 1 and piece[-1] == "," and piece[-2].isdigit():
            cur_pieces = sp_model.EncodeAsPieces(
                piece[:-1].replace(SPIECE_UNDERLINE, "")
            )
            if (
                piece[0] != SPIECE_UNDERLINE
                and cur_pieces[0][0] == SPIECE_UNDERLINE
            ):
                if len(cur_pieces[0]) == 1:
                    cur_pieces = cur_pieces[1:]
                else:
                    cur_pieces[0] = cur_pieces[0][1:]
            cur_pieces.append(piece[-1])
            new_pieces.extend(cur_pieces)
        else:
            new_pieces.append(piece)

    return new_pieces


def encode_ids(sp_model, text, sample=False):
    pieces = encode_pieces(sp_model, text, return_unicode=False, sample=sample)
    ids = [sp_model.PieceToId(piece) for piece in pieces]
    return ids


def load_vocab(vocab_file):
    """Loads a vocabulary file into a dictionary."""
    vocab = collections.OrderedDict()
    with open(vocab_file, "r", encoding="utf-8") as reader:
        tokens = reader.readlines()
    for index, token in enumerate(tokens):
        token = token.rstrip("\n")
        vocab[token] = index
    return vocab


def convert_by_vocab(vocab, items):
    """Converts a sequence of [tokens|ids] using the vocab."""
    output = []
    for item in items:
        output.append(vocab[item])
    return output


def convert_tokens_to_ids(vocab, tokens):
    return convert_by_vocab(vocab, tokens)


def convert_ids_to_tokens(inv_vocab, ids):
    return convert_by_vocab(inv_vocab, ids)


def whitespace_tokenize(text):
    """Runs basic whitespace cleaning and splitting on a piece of text."""
    text = text.strip()
    if not text:
        return []
    tokens = text.split()
    return tokens


def _is_whitespace(char):
    """Checks whether `chars` is a whitespace character."""
    # \t, \n, and \r are technically control characters but we treat them
    # as whitespace since they are generally considered as such.
    if char == " " or char == "\t" or char == "\n" or char == "\r":
        return True
    cat = unicodedata.category(char)
    if cat == "Zs":
        return True
    return False


def _is_control(char):
    """Checks whether `chars` is a control character."""
    # These are technically control characters but we count them as whitespace
    # characters.
    if char == "\t" or char == "\n" or char == "\r":
        return False
    cat = unicodedata.category(char)
    if cat in ("Cc", "Cf"):
        return True
    return False


def _is_punctuation(char):
    """Checks whether `chars` is a punctuation character."""
    cp = ord(char)
    # We treat all non-letter/number ASCII as punctuation.
    # Characters such as "^", "$", and "`" are not in the Unicode
    # Punctuation class but we treat them as punctuation anyways, for
    # consistency.
    if (
        (cp >= 33 and cp <= 47)
        or (cp >= 58 and cp <= 64)
        or (cp >= 91 and cp <= 96)
        or (cp >= 123 and cp <= 126)
    ):
        return True
    cat = unicodedata.category(char)
    if cat.startswith("P"):
        return True
    return False


from .AlbertTokenizer import (
    AlbertWordpieceTokenizer,
    AlbertBasicTokenizer,
    AlbertFullTokenizer,
)

from .HFAutoTokenizer import HFAutoTokenizer



from typing import Type
import torch
from collections import OrderedDict

# https://gist.github.com/the-bass/0bf8aaa302f9ba0d26798b11e4dd73e3
def rename_state_dict_keys(source, key_transformation, target=None):
    """Rename state dict keys

    Args:
        source             -> Path to the saved state dict.
        key_transformation -> Function that accepts the old key names of the state
                            dict as the only argument and returns the new key name.
        target (optional)  -> Path at which the new state dict should be saved
                            (defaults to `source`)
        Example:
        Rename the key `layer.0.weight` `layer.1.weight` and keep the names of all
        other keys.

    Example:
        ```
        def key_transformation(old_key):
            if old_key == "layer.0.weight":
                return "layer.1.weight"
            return old_key
        rename_state_dict_keys(state_dict_path, key_transformation)
        ```
    """
    if target is None:
        target = source

    state_dict = torch.load(source)
    new_state_dict = OrderedDict()

    for key, value in state_dict.items():
        new_key = key_transformation(key)
        new_state_dict[new_key] = value

    torch.save(new_state_dict, target)

from ednaml.utils.LabelMetadata import LabelMetadata
from ednaml.models.ModelAbstract import ModelAbstract
def build_model_and_load_weights(
    config_file: str,
    model_class: Type[ModelAbstract] = None,
    epoch: int = 0,
    custom_metadata: LabelMetadata = None,
    add_filehandler: bool = False,
    add_streamhandler: bool = True,
):
    """Generates a model using a configuration file, and loads a specific saved epoch

    Args:
        config_file (str): _description_
        model_class (Type[ModelAbstract], optional): _description_. Defaults to None.
        epoch (int, optional): _description_. Defaults to 0.

    Returns:
        _type_: _description_
    """
    from ednaml.core import EdnaML

    eml = EdnaML(
        config=config_file,
        mode="test",
        test_only=True,
        add_filehandler=add_filehandler,
        add_streamhandler=add_streamhandler,
    )
    eml.labelMetadata = custom_metadata  # TODO this needs to be fixed with the actual label metadata...or tell users to not infer things :|
    if model_class is not None:
        eml.addModelClass(model_class=model_class)
    eml.buildModel()
    if epoch is None:
        # means we get the most recent epoch available...
        epoch = eml.getPreviousStop()
    eml.loadEpoch(epoch=epoch)
    return eml.model


from typing import Any, Dict, List
from ednaml.config import BaseConfig

from ednaml.config.ConfigDefaults import ConfigDefaults


class TransformationConfig(BaseConfig):
    BATCH_SIZE: int
    WORKERS: int
    ARGS: Dict[str,Any]

    def __init__(self, transformation_dict, defaults: ConfigDefaults): #all of this will go, as shape mean etc won't matter
        self.BATCH_SIZE = transformation_dict.get(
            "BATCH_SIZE", defaults.BATCH_SIZE
        )
        self.WORKERS = transformation_dict.get("WORKERS", defaults.WORKERS)
        self.ARGS = transformation_dict.get("ARGS", defaults.TRANSFORM_ARGS)
        



from itertools import combinations
import numpy as np
import torch
import torch.nn.functional as F
from ednaml.loss import Loss


class CompactContrastiveLoss(Loss):
    """Standard contrastive loss

    Calculates the contrastive loss of a mini-batch.

    Args (kwargs only):
        positive-margin (float, 0.3): Margin constraint to use in loss. Positive classes MUST be closer than this value.
        negative-margin (float, 0.5): Negative classes must be further than d(x,p)+margin
        mine (str): Mining method. Default 'hard'. Supports ['hard', 'all'].

    Methods:
        __call__: Returns loss given features and labels.
    """

    def __init__(self, **kwargs):
        super(CompactContrastiveLoss, self).__init__()
        self.nmargin = kwargs.get("negative-margin", 0.5)
        self.pmargin = kwargs.get("positive-margin", 0.3)

    def forward(self, features, labels, epoch):
        positive_pairs, negative_pairs = self.get_pairs(features, labels)
        if features.is_cuda:
            positive_pairs = positive_pairs.cuda()
            negative_pairs = negative_pairs.cuda()
        if epoch % 2 == 0:
            loss = F.relu(
                (
                    features[positive_pairs[:, 0]]
                    - features[positive_pairs[:, 1]]
                )
                .pow(2)
                .sum(1)
                .sqrt()
                - self.pmargin
            ).pow(2)
        # (features[positive_pairs[:, 0]] - features[positive_pairs[:, 1]]).pow(2).sum(1)
        else:  # negative loss
            loss = F.relu(
                self.nmargin
                - (
                    features[negative_pairs[:, 0]]
                    - features[negative_pairs[:, 1]]
                )
                .pow(2)
                .sum(1)
                .sqrt()
            ).pow(2)
        # loss = torch.cat([positive_loss, negative_loss], dim=0)
        return loss.mean()

    def get_pairs(self, features, labels):
        distance_matrix = self.pdist(features)

        labels = labels.cpu().data.numpy()
        all_pairs = np.array(list(combinations(range(len(labels)), 2)))
        all_pairs = torch.LongTensor(all_pairs)
        positive_pairs = all_pairs[
            (labels[all_pairs[:, 0]] == labels[all_pairs[:, 1]]).nonzero()
        ]
        negative_pairs = all_pairs[
            (labels[all_pairs[:, 0]] != labels[all_pairs[:, 1]]).nonzero()
        ]

        negative_distances = distance_matrix[
            negative_pairs[:, 0], negative_pairs[:, 1]
        ]
        negative_distances = negative_distances.cpu().data.numpy()
        top_negatives = np.argpartition(
            negative_distances, len(positive_pairs)
        )[: len(positive_pairs)]
        top_negative_pairs = negative_pairs[torch.LongTensor(top_negatives)]

        return positive_pairs, top_negative_pairs

    def pdist(self, vectors):
        distance_matrix = (
            -2 * vectors.mm(torch.t(vectors))
            + vectors.pow(2).sum(dim=1).view(1, -1)
            + vectors.pow(2).sum(dim=1).view(-1, 1)
        )
        return distance_matrix



from typing import Any, Dict, Type

from ednaml.crawlers import Crawler
from ednaml.generators import Generator
import logging

class EdnaMLBase:
    logLevels = {
        0: logging.NOTSET,
        1: logging.ERROR,
        2: logging.INFO,
        3: logging.DEBUG,
    }
    _crawlerClassQueue: Type[Crawler]
    _crawlerArgsQueue: Dict[str, Any]
    _crawlerClassQueueFlag: bool
    _crawlerInstanceQueue: Crawler
    _crawlerInstanceQueueFlag: bool
    _generatorClassQueue = Type[Generator]
    _generatorArgsQueue = Dict[str, Any]
    _generatorClassQueueFlag = bool


from ednaml.core.EdnaML import EdnaML
from ednaml.core.EdnaDeploy import EdnaDeploy



from ednaml.config import BaseConfig
from ednaml.config.OptimizerConfig import OptimizerConfig


class OptimizerListConfig(BaseConfig):
    OPTIMIZER: OptimizerConfig

    def __init__(self):
        pass



from typing import List, Tuple, Union
from ednaml.generators import ImageGenerator
from torch.utils.data import Dataset as TorchDataset
from torch.utils.data import DataLoader as TorchDataLoader
from ednaml.utils import locate_class
from torchvision.datasets import VisionDataset
import torch
from ednaml.utils.LabelMetadata import LabelMetadata

class TorchvisionGeneratorWrapper(ImageGenerator):
    """Generator for a TorchVision dataset.

    Args:
        ImageGenerator (_type_): _description_
    """

    def buildGeneratorAttributes(self, **kwargs):
        self.torchvision_dataset_class = kwargs.get("tv_dataset")
        self.torchvision_dataset_args = kwargs.get("tv_args")

        
    def buildDataset(self, datacrawler, mode: str, transform: List[object], **kwargs) -> TorchDataset:
        # We don't need the data crawler...

        dataset_class: VisionDataset = locate_class(
            package="torchvision",
            subpackage="datasets",
            classpackage=self.torchvision_dataset_class,
        )
        # We probably need something here to adjust the arguments for each dataset class, because of differences in arguments...

        # dataset API is not unified, see discussions below
        # https://github.com/pytorch/vision/issues/1067
        # https://github.com/pytorch/vision/issues/1080
        if self.torchvision_dataset_class in ["CIFAR10", "MNIST", "USPS", "CIFAR100"]:
            # These have a train argument...
            return dataset_class(
                root=self.torchvision_dataset_args.get("root"),
                train=(mode == "train"),
                transform=self.transformer,
                target_transform=None,
                **self.torchvision_dataset_args.get("args", {})
            )
        elif self.torchvision_dataset_class in [
            "CelebA",
            "SVHN",
            "STL10",
            "ImageNet",
            "Cityscapes",
        ]:
            if self.torchvision_dataset_class == "ImageNet":
                mode = mode if mode == "train" else "val"
            return dataset_class(
                root=self.torchvision_dataset_args.get("root"),
                split=mode,
                transform=self.transformer,
                target_transform=None,
                **self.torchvision_dataset_args.get("args", {})
            )
        else:
            raise NotImplementedError()


    def buildDataLoader(self, dataset, mode, batch_size, **kwargs):
        return TorchDataLoader(
            dataset,
            batch_size=batch_size,
            shuffle=kwargs.get("shuffle", True),
            num_workers=self.workers,
        )

    def getNumEntities(self, datacrawler, mode, **kwargs) -> LabelMetadata:
        label_name = kwargs.get("label_name", "label")
        num_classes = kwargs.get("num_classes", None)
        if num_classes is None:
            num_classes = torch.unique(torch.Tensor(self.dataset.targets)).shape[0]
        return LabelMetadata({label_name: {"classes": num_classes}})


from ednaml.crawlers import Crawler
from ednaml.datareaders import DataReader
from ednaml.generators.TorchvisionGeneratorWrapper import (
    TorchvisionGeneratorWrapper,
)


class TorchvisionDatareader(DataReader):
    name: str = "TorchvisionDatareader"
    GENERATOR = TorchvisionGeneratorWrapper
